
// $Id: AlienCommand.cpp,v 1.18 2014/02/05 08:36:53 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AlienCommand.h"

//___________________________________________________________________
unsigned int& AlienCommand::_counter( void )
{
    static unsigned int counter( 0 );
    return counter;
}

//___________________________________________________________________
AlienCommand::AlienCommand( const Type type ):
    Counter( "AlienCommand" ),
    id_( _counter()++ ),
    priority_( 0 ),
    type_( type )
{}

//___________________________________________________________________
QString AlienCommand::name( const Type& type )
{
    Debug::Throw( "AlienCommand::name.\n" );

    if( _names().empty() ) _initializeNames();
    NameMap::const_iterator iter( _names().find( type ) );
    return iter != _names().end() ? iter.value():"";
}

//___________________________________________________________________
QString AlienCommand::user( void ) const
{
    switch( type_ )
    {
        case DefaultMasterList:
        case MasterList:
        case ChildrenList:
        case Kill:
        case KillMaster:
        case Resubmit:
        case ResubmitMaster:
        case Trace:
        case TraceAll:
        case TraceJdl:
        case ClearQueue:
        return arguments_[0];

        default: return QString();
    }

}

//___________________________________________________________________
int AlienCommand::parentJobId( void ) const
{
    switch( type_ )
    {
        case ChildrenList:
        case Resubmit:
        case Trace:
        case TraceAll:
        case TraceJdl:
        case JobFiles:
        return arguments_[1].toInt();
        default: return -1;
    }
}

//___________________________________________________________________
int AlienCommand::jobId( void ) const
{
    switch( type_ )
    {
        case ChildrenList:
        case Kill:
        case KillMaster:
        case Resubmit:
        case ResubmitMaster:
        case Trace:
        case TraceAll:
        case TraceJdl:
        case JobFiles:
        return arguments_[2].toInt();
        default: return -1;
    }
}

//___________________________________________________________________
QString AlienCommand::path( void ) const
{
    switch( type_ )
    {
        case JobFiles: return arguments_[3];
        default: return QString();
    }
}

//___________________________________________________________________
void AlienCommand::_initializeNames( void )
{
    Debug::Throw( "AlienCommand::_initializeNames.\n" );

    _names()[None] = "none";
    _names()[Connect] = QObject::tr( "connect to the grid" );
    _names()[CheckProxy] = QObject::tr( "check GSI proxy validity" );
    _names()[FlushErrors] = QObject::tr( "check failed commands" );
    _names()[Copy]= QObject::tr( "copy selected file" );
    _names()[Remove]= QObject::tr( "remove selected file" );
    _names()[RemoveDirectory]= QObject::tr( "remove selected directory" );
    _names()[DefaultUser]=QObject::tr( "query default user name" );
    _names()[DefaultMasterList]=QObject::tr( "query master job list for default user" );
    _names()[MasterList]=QObject::tr( "query master job list for all users" );
    _names()[ChildrenList]=QObject::tr( "query job children list" );
    _names()[Kill]=QObject::tr( "kill selected job(s)" );
    _names()[KillMaster]=QObject::tr( "kill selected master job(s)" );
    _names()[Resubmit]=QObject::tr( "resubmit selected job(s)" );
    _names()[ResubmitMaster]=QObject::tr( "resubmit selected job(s)" );
    _names()[Trace]=QObject::tr( "query job information" );
    _names()[TraceJdl]=QObject::tr( "query job JDL" );
    _names()[TraceAll]=QObject::tr( "query job information" );
    _names()[JobFiles]=QObject::tr( "query job output files" );
    _names()[ClearQueue]=QObject::tr( "clear command queue" );
    _names()[UserList]=QObject::tr( "query user list" );

}

//___________________________________________________________________
QString AlienCommand::errorString( void ) const
{
    switch( type_ )
    {
        case CheckProxy:
        return QObject::tr( "No valid GSI proxy was found. "
            "All connections to the grid are likely to fail. "
            "Use <xrdgsiproxy init> to get a valid proxy." );

        case Connect:
        return QObject::tr( "Failed to connect to remote file system." );

        case Copy: return QString( QObject::tr( "Failed to copy file '%1'" ) ).arg( arguments_[0] );
        case Remove: return QString( QObject::tr( "Failed to delete file '%1'" ) ).arg( arguments_[0] );
        case RemoveDirectory: return QString( QObject::tr( "Failed to delete folder '%1'" ) ).arg( arguments_[0] );
        case ChildrenList: return QString( QObject::tr( "Failed to query children list for job %1" ) ).arg( jobId() );
        case Kill: return QString( QObject::tr( "Failed to kill job %1" ) ).arg( jobId() );
        case KillMaster: return QString( QObject::tr( "Failed to kill job %1" ) ).arg( jobId() );
        case Resubmit: return QString( QObject::tr( "Failed to resubmit job %1" ) ).arg( jobId() );
        case ResubmitMaster: return QString( QObject::tr( "Failed to resubmit failed children for job %1" ) ).arg( jobId() );
        case Trace: return QString( QObject::tr( "Failed to retrieve information for job %1" ) ).arg( jobId() );
        case TraceAll: return QString( QObject::tr( "Failed to query information for job %1" ) ).arg( jobId() );
        case TraceJdl: return QString( QObject::tr( "Failed to query JDL information for job %1" ) ).arg( jobId() );
        case JobFiles: return QString( QObject::tr( "Failed to query output files for job %1 (path: '%2')" ) ).arg( jobId() ).arg( path() );
        default: return QString( QObject::tr( "Failed to %1" ) ).arg( name() );
    }

}

//___________________________________________________________________
AlienCommand::NameMap& AlienCommand::_names( void )
{
    static NameMap names;
    return names;
}
