// $Id: AlienInterface.cpp,v 1.61 2014/03/21 11:50:54 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AlienInterface.h"
#include "AlienInterface.moc"

#include "Application.h"
#include "CustomProcess.h"
#include "Command.h"
#include "Debug.h"
#include "File.h"
#include "FileRecordProperties.h"
#include "IOString.h"
#include "JobSet.h"
#include "QOrderedSet.h"
#include "Singleton.h"
#include "TextFormat.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QRegExp>
#include <QTextStream>

#include <algorithm>

//_______________________________________________
AlienInterface::AlienInterface( QObject *parent ):
    QObject( parent ),
    Counter( "AlienInterface" ),
    paused_( false ),
    hideKilledJobs_( false ),
    process_( 0 ),
    wasConnected_( false ),
    connected_( true ),
    ignoreErrors_( false )
{
    Debug::Throw("AlienInterface::AlienInterface.\n");
    connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();
}

//_______________________________________________
AlienInterface::~AlienInterface( void )
{ Debug::Throw("AlienInterface::~AlienInterface.\n"); }

//_______________________________________________
void AlienInterface::findCertificates( void )
{

    Debug::Throw( "AlienInterface::findCertificates.\n" );

    // check if option is already set and valid
    if( XmlOptions::get().contains( "CERTIFICATES_PATH" ) &&
        File( XmlOptions::get().raw( "CERTIFICATES_PATH" ) ).exists() &&
        File( XmlOptions::get().raw( "CERTIFICATES_PATH" ) ).isAbsolute() )
    { return; }

    File alienPath;

    // try load from gbbox path
    if( XmlOptions::get().contains( "GBBOX" ) && File( XmlOptions::get().raw( "GBBOX" ) ).isAbsolute() )
    {
        File gbboxPath( File( XmlOptions::get().raw( "GBBOX" ) ).path() );
        alienPath = File( "../.." ).addPath( gbboxPath ).canonicalName();
    }

    if( !alienPath.exists() )
    {

        // try load using GSHELL_ROOT
        const QProcessEnvironment environment( QProcessEnvironment::systemEnvironment() );
        if( environment.contains( "GSHELL_ROOT" ) )
        { alienPath = File( ".." ).addPath( environment.value( "GSHELL_ROOT" ) ).canonicalName(); }

        // try load using ALIEN_ROOT environment
        if( !alienPath.exists() && environment.contains( "ALIEN_ROOT" ) )
        { alienPath = File( environment.value( "ALIEN_ROOT" ) ).canonicalName(); }

    }

    if( !alienPath.exists() ) return;

    // possible path for certificates, relative to alien path
    File::List pathList;
    pathList
        << File( "globus/share/certificates" )
        << File( "api/share/certificates" );

    // get certificates path
    File certificatesPath;
    foreach( const File& path, pathList )
    {
        const File tmp( path.addPath( alienPath ).canonicalName() );
        if( tmp.exists() )
        {
            certificatesPath = tmp;
            break;
        }
    }

    // assign
    if( certificatesPath.exists() )
    { XmlOptions::get().setRaw( "CERTIFICATES_PATH", certificatesPath ); }

    return;

}

//__________________________________________________________________________________
void AlienInterface::addCommand( const AlienCommand& command )
{

    // keep empty state
    const bool empty( commands_.isEmpty() );

    // doe nothing if command is not effectively added (duplicate)
    emit commandAdded( command );
    if( !commands_.append( command ) )
    {
        emit commandCompleted( command );
        return;
    }

    Debug::Throw() << "AlienInterface::addCommand - " << command.name() << " (" << command.id() << ")" << endl;


    // emit signal

    // run command if there is none pending
    if( empty )
    {
        emit busyStateChanged( true );
        emit busy();
        _runCommand();
    }
}

//__________________________________________________________________________________
void AlienInterface::addCommands( const AlienCommand::List& commands )
{

    Debug::Throw( "AlienInterface::addCommands.\n" );

    // list of added commands
    AlienCommand::List addedCommands;

    const bool empty( commands_.isEmpty() );
    foreach( const AlienCommand& command, commands )
    {
        if( commands_.append( command ) )
        {

            Debug::Throw() << "AlienInterface::addCommands - " << command.name() << " (" << command.id() << ")" << endl;
            addedCommands << command;

        } else {

            emit commandAdded( command );
            emit commandCompleted( command );

        }

    }

    // check if commands have been effectively added
    if( addedCommands.isEmpty() ) return;

    emit commandsAdded( addedCommands );
    if( empty )
    {
        emit busyStateChanged( true );
        emit busy();
        _runCommand();
    }

}

//__________________________________________________________________________________
void AlienInterface::abortAll( void )
{

    Debug::Throw( "AlienInterface::abortAll.\n" );

    // abort all commands but the current
    foreach( const AlienCommand& command, commands_.commands() )
    { if( commands_.abort( command ) ) emit commandCompleted( command ); }

    // kill process if any. this should also abort current command
    if( process_ && process_->state() != QProcess::NotRunning )
    { process_->kill(); }

}

//__________________________________________________________________________________
void AlienInterface::abortCommands( const AlienCommand::List& commands )
{
    Debug::Throw( "AlienInterface::abortCommands.\n" );
    foreach( const AlienCommand& command, commands )
    { if( commands_.abort( command ) ) emit commandCompleted( command ); }
}

//____________________________________________________________
void AlienInterface::clearQueue( const QString& user, const Job::IdSet& jobs )
{

    Debug::Throw() << "Alieninterface::clearQueue - user: " << user << " jobs: " << jobs.size() << endl;

    // list of commands to be aborted.
    AlienCommand::List abortedCommands;

    // get all commands from queue
    SortedAlienCommand::Set commands( commands_.commands() );
    foreach( const SortedAlienCommand& command, commands )
    {
        // check user
        if( command.user() != user ) continue;

        // check job and parent id, for removal
        if(
            ( command.jobId() >= 0 && !jobs.contains( command.jobId() ) ) ||
            ( command.parentJobId() >= 0 && !jobs.contains( command.parentJobId() ) ) )
        { abortedCommands << command; }

    }

    if( !abortedCommands.isEmpty() )
    { abortCommands( abortedCommands ); }

}

//__________________________________________________________________________________
bool AlienInterface::_runCommand( void )
{

    // do nothing if paused
    if( paused_ ) return false;

    // do nothing if no commands are in the pile
    if( commands_.isEmpty() )
    {

        Debug::Throw( "AlienInterface::_runCommand - ready \n" );

        QString buffer = QString( tr( "Last update: %1" ) ).arg( TimeStamp::now().toString( TimeStamp::TimeLong ) );
        emit messageAvailable( buffer );

        QMetaObject::invokeMethod( this, "ready", Qt::QueuedConnection );
        emit busyStateChanged( false );

        return false;

    }

    // front command
    AlienCommand command( commands_.front() );
    Debug::Throw() << "AlienInterface::_runCommand - command: " << command.name() << " (" << command.id() << ")" << endl;

    // reset ignore error flag
    if( command.type() != AlienCommand::FlushErrors )
    { ignoreErrors_ = false; }

    /*
    don't run command if it is associated to a user that is not in the
    list of active users any more
    */
    if(
        (
        command.type() == AlienCommand::DefaultMasterList ||
        command.type() == AlienCommand::MasterList ||
        command.type() == AlienCommand::ChildrenList ||
        command.type() == AlienCommand::Trace ) &&
        !activeUsers_.contains( command.user() ) )
    {
        emit commandCompleted( command );
        commands_.removeFirst();
        return _runCommand();
    }

    // reset buffer
    _resetBuffers();
    _resetProcess();

    // take action depending on command type
    switch( command.type() )
    {

        case AlienCommand::CheckProxy:
        {

            emit logAvailable( tr( "Checking GSI proxy\n" ), Format::Bold );
            QStringList arguments;
            arguments
                << XmlOptions::get().raw("GSI_PROXY" )
                << "info";

            _newProcess()->start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::Connect:
        {

            emit logAvailable( tr( "Trying to connect to remote filesystem\n" ), Format::Bold );
            QStringList arguments;
            arguments
                << XmlOptions::get().raw("GBBOX" )
                << "connect"
                << XmlOptions::get().raw( "ALIEN_HOST" )
                << QString::number( XmlOptions::get().get<int>( "ALIEN_PORT" ) )
                << XmlOptions::get().raw( "ALIEN_USER" );

            _newProcess()->start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::FlushErrors:
        {

            if( ignoreErrors_ ) clearFailedCommands();
            else {
                if( !failedCommands_.isEmpty() )  emit commandsFailed( failedCommands_ );
                if( !failedCommands_.isEmpty() )
                {
                    emit logAvailable( tr( "Trying to recover failed commands\n" ), Format::Bold );
                    const AlienCommand::List commands( failedCommands_ );
                    clearFailedCommands();
                    addCommands( commands );
                    addCommand( AlienCommand::FlushErrors );

                }

            }

            break;

        }

        case AlienCommand::DefaultUser:
        {

            emit logAvailable( tr( "Requesting default user\n" ), Format::Bold );
            emit messageAvailable( tr( "Requesting default user" ) );

            QStringList arguments;
            arguments
                << XmlOptions::get().raw("GBBOX" )
                << "whoami";

            // run and emit signal
            _newProcess()->start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::UserList:
        {

            emit logAvailable( tr( "Requesting user list\n" ), Format::Bold );
            emit messageAvailable( tr( "Requesting user list" ) );

            QStringList arguments;
            arguments
                << XmlOptions::get().raw("GBBOX" )
                << "who";

            // run and emit signal
            _newProcess()->start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::Copy:
        {

            const QString remote( command.arguments()[0] );
            const QString local( command.arguments()[1] );

            // emit message
            QString buffer = QString( tr( "Copying file %1" ) ).arg( remote );
            emit logAvailable( buffer + "\n", Format::Bold );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments
                << XmlOptions::get().raw("ALIEN_CP")
                << QString( "alien:" ) + remote
                << QString( "file:" ) + local;

            // run and emit signal
            _newProcess()->start( arguments );
            emit commandStarted( command );

            return true;
        }

        case AlienCommand::Remove:
        {
            const QString remote( command.arguments()[0] );

            // emit message
            QString buffer = QString( tr( "Removing file %1" ) ).arg( remote );
            emit logAvailable( buffer + "\n", Format::Bold );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments
                << XmlOptions::get().raw("GBBOX")
                << QString( "rm" ) << remote;

            // run and emit signal
            _newProcess()->start( arguments );
            emit commandStarted( command );

            return true;
        }

        case AlienCommand::RemoveDirectory:
        {
            const QString remote( command.arguments()[0] );

            // emit message
            QString buffer = QString( tr( "Removing folder %1" ) ).arg( remote );
            emit logAvailable( buffer + "\n", Format::Bold );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments
                << XmlOptions::get().raw("GBBOX")
                << QString( "rmdir" ) << remote;

            // run and emit signal
            _newProcess()->start( arguments );
            emit commandStarted( command );

            return true;
        }

        case AlienCommand::DefaultMasterList:
        {

            // get user and check
            const QString user( command.user() );
            if( user.isEmpty() ) break;

            // emit message
            QString buffer = QString( tr( "Requesting job list for user %1" ) ).arg( user );
            emit logAvailable( buffer + "\n", Format::Bold );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments << XmlOptions::get().raw("GBBOX" ) << "ps" << "-A";

            // run and emit signal
            _newProcess()->start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::MasterList:
        {

            // get user and check
            QStringList users( command.arguments() );
            if( users.isEmpty() ) break;

            // emit message
            QString buffer = (users.size() == 1) ?
                QString( tr( "Requesting job list for user %1" ) ).arg( users.front() ):
                QString( tr( "Requesting job list for users %1" ) ).arg( users.join( " " ) );
            emit logAvailable( buffer + "\n", Format::Bold );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments << XmlOptions::get().raw("GBBOX" ) << "ps" << "-A" << "-a";

            // run and emit signal
            _newProcess()->start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::ChildrenList:
        {

            // check arguments
            const QString user( command.user() );
            const int id( command.jobId() );

            // emit message
            QString buffer = QString( tr( "Requesting children list for job %1" ) ).arg( id );
            emit logAvailable( buffer + "\n", Format::Bold );
            emit messageAvailable( buffer );

            // create command
            QStringList arguments;
            arguments
                << XmlOptions::get().raw("GBBOX") << "top"
                << "-user" << user
                << "-split" << QString::number( id )
                << "-all_status";

            // run and emit signal
            _newProcess()->start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::Kill:
        {

            // get user and check
            const int id( command.jobId() );

            // emit message
            QString buffer = QString( tr( "Killing job %1" ) ).arg( id );
            emit logAvailable( buffer + "\n", Format::Bold );
            emit messageAvailable( buffer );

            QStringList arguments;
            arguments << XmlOptions::get().raw( "GBBOX" ) << "kill" << QString::number( id );

            // run and emit signal
            _newProcess()->start( arguments );
            emit commandStarted( command );
            return true;
        }

        case AlienCommand::KillMaster:
        {

            // get id
            const int id( command.jobId() );
            if( id < 0 ) break;

            // emit message
            QString buffer = QString( tr( "Killing job %1" ) ).arg( id );
            emit logAvailable( buffer + "\n", Format::Bold );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments << XmlOptions::get().raw( "GBBOX" ) << "masterJob" << QString::number(id);

            // add status
            if( command.arguments().size() >= 4 )
            { arguments << "-status" << command.arguments()[3]; }

            arguments << "kill";

            // run and emit signal
            _newProcess()->start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::Resubmit:
        {

            // get user and check
            const int id( command.jobId() );
            if( id < 0 ) break;

            // emit message
            QString buffer = QString( tr( "Resubmitting job %1" ) ).arg( id );
            emit logAvailable( buffer + "\n", Format::Bold );
            emit messageAvailable( buffer );

            QStringList arguments;
            arguments << XmlOptions::get().raw( "ALIEN_RESUBMIT" ) << QString::number( id );

            // run and emit signal
            _newProcess()->start( arguments );
            emit commandStarted( command );
            return true;
        }

        case AlienCommand::ResubmitMaster:
        {

            // check arguments
            if( !command.hasArguments() ) break;

            // get user and check
            const int id( command.jobId() );
            if( id < 0 ) break;

            // get status from arguments, if any
            const QString status( command.arguments().size() >= 4 ? command.arguments()[3]:"ERROR_ALL" );

            // run and emit signal
            QString buffer = QString( tr( "Resubmitting failed jobs for job %1 (status: %2)" ) ).arg( id ).arg( status );
            emit logAvailable( buffer + "\n", Format::Bold );
            emit messageAvailable( buffer );

            QStringList arguments;
            arguments << XmlOptions::get().raw( "GBBOX" ) << "masterJob" << QString::number( id ) << "-status" << status << "resubmit";

            // start process
            _newProcess()->start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::Trace:
        case AlienCommand::TraceAll:
        {

            // get id
            const int parentId( command.parentJobId() );
            const int id( command.jobId() );

            // emit message
            QString buffer = QString( tr( "Requesting info for job %1" ) ).arg( id );
            emit logAvailable( buffer + "\n", Format::Bold );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments
                << XmlOptions::get().raw("ALIEN_PS")
                << "-trace"
                << QString::number( id );

            // add output selection string based on command type
            if( command.type() == AlienCommand::Trace )
            {

                // request all state lines
                arguments << "state";

                // all so request "submission" lines, to get list of children
                if( parentId < 0 ) arguments << "submit";

            } else if( command.type() == AlienCommand::TraceAll ) arguments << "all";

            // run and emit signal
            _newProcess()->start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::TraceJdl:
        {

            // get id
            const int id( command.jobId() );

            // emit message
            QString buffer = QString( tr( "Requesting JDL for job %1" ) ).arg( id );
            emit logAvailable( buffer + "\n", Format::Bold );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments
                << XmlOptions::get().raw("ALIEN_PS")
                << "-jdl"
                << QString::number( id );

            // run and emit signal
            _newProcess()->start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::JobFiles:
        {

            // get id
            const int id( command.jobId() );
            const QString path( command.path() );

            // emit message
            QString buffer = QString( tr( "Requesting output files list for job %1" ) ).arg( id );
            emit logAvailable( buffer + "\n", Format::Bold );
            emit messageAvailable( buffer );

            // build command
            QStringList arguments;
            arguments
                << XmlOptions::get().raw( "GBBOX" )
                << "ls" << "-n" << "-l" << path;

            // run and emit signal
            _newProcess()->start( arguments );
            emit commandStarted( command );
            return true;

        }

        case AlienCommand::ClearQueue:
        {

            const QString user( command.user() );
            Q_ASSERT( !user.isEmpty() );

            // emit message
            QString buffer = QString( tr( "Clearing command queue for user %1" ) ).arg( user );
            emit logAvailable( buffer + "\n", Format::Bold );
            emit messageAvailable( buffer );
            break;
        }

        default:
        break;
    }

    emit commandCompleted( command );
    commands_.removeFirst();
    return _runCommand();

}

//______________________________________________________________
void AlienInterface::_commandCompleted( int returnCode, QProcess::ExitStatus status )
{

    /*
    if there is no current command in the queue (in case of abort)
    just jump to the next macro
    */
    if( commands_.isEmpty() )
    {
        _runCommand();
        return;
    }

    // fetch command
    AlienCommand command( commands_.front() );
    Debug::Throw()
        << "AlienInterface::_commandCompleted -"
        << " command: " << command.name() << " (" << command.id() << ")"
        << " status: " << status
        << " code: " << returnCode << endl;

    // check connection state
    const bool ignoreDisconnect(
        command.type() == AlienCommand::CheckProxy ||
        command.type() == AlienCommand::FlushErrors );
    const bool connected( ignoreDisconnect || _checkConnection() );

    // force return code to zero for MasterList
    if(
        (
        command.type() == AlienCommand::DefaultMasterList ||
        command.type() == AlienCommand::MasterList ||
        command.type() == AlienCommand::ChildrenList ||
        command.type() == AlienCommand::UserList )
        && returnCode == 1 ) returnCode = 0;

    // check status
    if( status != QProcess::NormalExit || ( process_ && process_->error() == QProcess::FailedToStart ) || returnCode != 0 || !connected )
    {

        const QString buffer = command.errorString();
        emit messageAvailable( buffer );
        emit logAvailable( buffer + "\n\n", Format::Italic|Format::Bold );

        // store failed command
        command.setErrorDetails( stderrBuffer_.text() );
        failedCommands_.append( command );

        emit commandCompleted( command );
        commands_.removeFirst();

        // add FlushError command
        if( !_ignoreError( command, stderrBuffer_.text() ) )
        { addCommand( AlienCommand::FlushErrors ); }

        _runCommand();
        return;

    }

    // add (cosmetics) empty line to log.
    emit logAvailable( "\n" );

    // run action depending on current type
    switch( command.type() )
    {

        case AlienCommand::CheckProxy:
        {
            _processProxyInfo();
            break;
        }

        case AlienCommand::DefaultUser:
        {

            _processDefaultUser();
            break;
        }

        case AlienCommand::UserList:
        {

            _processUserList();
            break;
        }

        case AlienCommand::Copy:
        {

            // run external application, if provided
            if( command.arguments().size() >= 3 )
            { ( Command( command.arguments()[2] ) << command.arguments()[1] ).run(); }

            break;
        }

        case AlienCommand::DefaultMasterList:
        case AlienCommand::MasterList:
        {
            _processMasterList( command.arguments() );
            break;
        }

        case AlienCommand::ChildrenList:
        {
            Job job;
            job.assign( Job::User, command.user() )
                .assign( Job::Id, command.jobId() );
            _processChildrenList( job );
            break;
        }

        case AlienCommand::Trace:
        case AlienCommand::TraceAll:
        {
            Job job;
            job.setParentId( command.parentJobId() )
                .assign( Job::User, command.user() )
                .assign( Job::Id, command.jobId() );
            _processJobInformation( job, command.type() == AlienCommand::TraceAll );
            break;
        }

        case AlienCommand::TraceJdl:
        {
            Job job;
            job.setParentId( command.parentJobId() )
                .assign( Job::User, command.user() )
                .assign( Job::Id, command.jobId() );
            _processJdl( job );
            break;
        }

        case AlienCommand::JobFiles:
        {
            Job job;
            job.setParentId( command.parentJobId() )
                .assign( Job::User, command.user() )
                .assign( Job::Id, command.jobId() );
            _processJobFiles( job, command.arguments()[3] );
             break;
        }

        default: break;

    }

    // emit signal, remove from pile and run next command if any
    emit commandCompleted( command );
    commands_.removeFirst();
    _runCommand();

    return;

}

//____________________________________________________________
void AlienInterface::_error( QProcess::ProcessError error )
{

    Q_ASSERT( !commands_.isEmpty() );
    const AlienCommand command( commands_.front() );
    Debug::Throw() << "AlienInterface::_error - command: " << command.name() << " (" << command.id() << ")" << endl;

    if( error == QProcess::FailedToStart )
    {

        /*
        in case the process failed to start, _commandComppleted is not called
        one needs to call it manually in order to proceed with command list
        */
        _commandCompleted( 1, QProcess::NormalExit );

    } else {

        const QString buffer = command.errorString();
        emit messageAvailable( buffer );
        emit logAvailable( buffer+"\n\n", Format::Italic|Format::Bold );

    }

    return;

}

//________________________________________________________________________
bool AlienInterface::_checkConnection( void )
{

    Debug::Throw() << "AlienInterface::_checkConnection -"
        << " connected: " << connected_
        << " wasConnected: " << wasConnected_
        << endl;

    // Warning [Mon Oct 29 09:39:48 2012] GCLIENT::Reconnect
    // Warning [Mon Oct 29 09:39:48 2012] GCLIENT::COMMAND command failed
    QRegExp regExp1( "(Error)|(Warning)\\s+\\[.+\\]\\s+GCLIENT::Reconnect" );
    QRegExp regExp2( "(Error)|(Warning)\\s+\\[.+\\]\\s+GCLIENT::COMMAND command failed" );
    if( stderrBuffer_.text().indexOf( regExp1 ) >= 0 && stderrBuffer_.text().indexOf( regExp2 ) >= 0 )
    {

        if( connected_ )
        {
            connected_ = false;
            emit connectionLost( stderrBuffer_.text() );
            emit connectionStateChanged( false );
        }

        return false;

    } else if( !( connected_ && wasConnected_ ) ) {

        emit connectionStateChanged( true );
        wasConnected_ = true;
        connected_ = true;

    }

    return true;

}

//________________________________________________________________________
void AlienInterface::_processProxyInfo( void )
{
    Debug::Throw( "AlienInterface::_processProxyInfo.\n" );

    // get current command
    AlienCommand command( commands_.front() );
    Q_ASSERT( command.type() == AlienCommand::CheckProxy );

    // proxy validity
    bool valid( false );

    // time left   : 6h:52m:40s
    QRegExp regexp( "time\\s+left\\s+:\\s+(\\d+)h:(\\d+)m:(\\d+)s" );
    if( stderrBuffer_.text().indexOf( regexp ) >= 0 )
    {

        const int totalTime = 3600*regexp.cap(1).toInt() + 60*regexp.cap(2).toInt() + regexp.cap(3).toInt();
        valid = ( totalTime > 0 );

    }

    // if invalid emit signal
    if( !valid )
    {
        command.setErrorDetails( stderrBuffer_.text() );
        failedCommands_.append( command );
        addCommand( AlienCommand::FlushErrors );
    }

}

//________________________________________________________________________
void AlienInterface::_processDefaultUser( void )
{

    Debug::Throw( "AlienInterface::_processDefaultUser.\n" );
    QString defaultUser;

    // look for default user first in stdout then stderr
    foreach( QString buffer, QStringList() << stdoutBuffer_.text() << stderrBuffer_.text() )
    {
        QTextStream stream( &buffer, QIODevice::ReadOnly );
        QStringList lines;
        QString line;
        while( !( line = stream.readLine() ).isNull() ) lines << line;

        // read default user
        if( !lines.isEmpty() && !_isError( line = lines.back() ) )
        { defaultUser = line.trimmed(); }

        if( !defaultUser.isEmpty() ) break;
    }

    // build path from user name
    if( defaultUser.isEmpty() )
    {

        emit messageAvailable( tr( "Failed to retrieve default user" ) );

    } else {

        XmlOptions::get().setRaw( "ALIEN_USER", defaultUser );
        emit messageAvailable( "Default user available" );
        emit defaultUserAvailable( defaultUser );

    }

}

//________________________________________________________________________
void AlienInterface::_processUserList( void )
{

    Debug::Throw( "AlienInterface::_processUserList.\n" );
    UserSet users;
    users.setNeedsDefault( false );

    // parse output
    QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
    QString line;

    QOrderedSet<QString> orderedUsers;
    while( !( line = stream.readLine() ).isNull() )
    { if( !( line = line.trimmed() ).isEmpty() ) orderedUsers.insert( line ); }

    foreach( const QString& user, orderedUsers )
    { users << user; }

    emit messageAvailable( "User list available" );
    emit userListAvailable( users );

}

//________________________________________________________________________
void AlienInterface::_processMasterList( const QStringList& users )
{

    Debug::Throw( "AlienInterface::_processMasterList.\n" );

    // list of found jobs
    JobMap jobs;
    JobMap killedJobs;

    // generate one empty list per user
    foreach( const QString& user, users )
    { jobs.insert( user, JobSet() ); }

    // parse output
    QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
    QString line;
    while( !( line = stream.readLine() ).isNull() )
    {

        if( line.isEmpty() ) continue;

        // create job from line
        Job job(line.trimmed());

        // check validity
        if( !job.isValid() ) continue;

        // check user
        const QString user( job.get( Job::User ) );
        if( !users.contains( user ) ) continue;

        // store job in relevant job sets
        if( job.isKilled() )
        {

            killedJobs[user].insert( job );
            if( !hideKilledJobs_ ) jobs[user].insert( job );

        } else {

            // append to job sest
            jobs[user].insert( job );

        }

    }

    // store killed jobs
    for( JobMap::const_iterator iter = killedJobs.constBegin(); iter != killedJobs.constEnd(); ++iter )
    { killedJobs_[iter.key()] = iter.value(); }

    emit messageAvailable( "Master job list available" );

    for( JobMap::const_iterator iter = jobs.constBegin(); iter != jobs.constEnd(); ++iter )
    { emit masterListAvailable( iter.key(), iter.value() ); }

}

//________________________________________________________________________
void AlienInterface::_processChildrenList( const Job& master )
{

    Debug::Throw( "AlienInterface::_processJobList.\n" );

    // list of found jobs
    JobSet jobs;
    JobSet killedJobs;

    // get user
    const QString user( master.get( Job::User ) );

    // parse output
    QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
    QString line;
    while( !( line = stream.readLine() ).isNull() )
    {

        if( line.isEmpty() ) continue;

        // create job from line
        Job job(line.trimmed());

        // assign user and parent
        job.assign( Job::User, user );
        job.setParentId( master.id() );

        // check validity
        if( !job.isValid() )
        {
            Debug::Throw() << "AlienInterface::_processChildrenList - invalid job: " << line << endl;
            continue;
        }

        Debug::Throw() << "AlienInterface::_processChildrenList - user: " << job.get( Job::User ) << " parent:" << job.parentId() << " id: " << job.id() << endl;

        // store job in relevant job sets
        if( job.isKilled() )
        {

            killedJobs.insert( job );
            if( !hideKilledJobs_ ) jobs.insert( job );

        } else  jobs.insert( job );

    }

    // store killed jobs
    killedJobs_[user] = killedJobs;

    emit childrenListAvailable( master, jobs );

}

//________________________________________________________________________
void AlienInterface::_processJobFiles( const Job& job, const QString& path )
{
    Debug::Throw( "AlienInterface::_processJobFiles" );

    // list of files
    FileRecord::List files;

    // regular expression
    QRegExp fileRegExp(
        "(\\S+)\\s+"                                      // permissions
        "(\\w+)\\s+(\\w+)\\s+"                            // user and group
        "(\\d+)\\s+"                                      // size
        "((?:\\w+)\\s+(?:\\d+)\\s+(?:\\d+):(?:\\d+))\\s+" // date
        "(\\S+)" );

    QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
    QString line;
    while( !( line = stream.readLine() ).isNull() )
    {

        // parse line
        if( fileRegExp.indexIn(line) < 0 ) continue;

        FileRecord file;
        file.setFile( File( fileRegExp.cap(6) ).addPath( path ) );
        file.addProperty( FileRecordProperties::Permissions, fileRegExp.cap(1) );
        file.addProperty( FileRecordProperties::User, fileRegExp.cap(2) );
        file.addProperty( FileRecordProperties::Group, fileRegExp.cap(3) );
        file.addProperty( FileRecordProperties::Size, fileRegExp.cap(4) );
        file.addProperty( FileRecordProperties::Creation, fileRegExp.cap(5) );

        files << file;

    }

    emit jobFilesAvailable( job, path, files );

}

//________________________________________________________________________
void AlienInterface::_processJobInformation( Job job, bool emitTrace )
{

    Debug::Throw( "AlienInterface::_processJobInformation.\n" );

    // regular expression to match data
    // date string is: Tue Nov  1 15:49:06 2011
    QRegExp timeRegExp( "(\\w+)\\s+(\\w+)\\s+(\\d+)\\s+(\\d+):(\\d+):(\\d+)\\s+(\\d+)\\s+" );

    // parent ID regExp  [Master Job is 135064557]
    QRegExp parentRegExp( "\\[Master Job is (\\d+)\\]" );

    // child ID regExp
    QRegExp childRegExp( "Subjob submitted:\\s+(\\d+)" );
    QRegExp resubmissionRegExp( "resubmiting subjob\\s+(\\d+)\\s+\\((\\d+)\\)" );
    QRegExp newJobRegExp( "New job \\(resubmission from (\\d+)\\)" );

    // line type enumeration
    enum LineType
    {
        None = 0,
        Child = 1<<0,
        Parent = 1<<1,
        Resubmission = 1<<2,
        NewJob = 1<<3,
        Starting = 1<<4,
        Running = 1<<5,
        Completed = 1<<6
    };

    // children job set
    JobSet children;

    // id of resubmitted job
    int resubmitted( -1 );

    // loop over lines, and look for trace information
    bool first( true );

    // parse output
    QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
    QString line;
    while( !( line = stream.readLine() ).isNull() )
    {

        // decide line type
        unsigned int lineType( None );

        // child job
        Job child;

        if( line.indexOf( "[submit ]" ) >= 0 )
        {

            if( childRegExp.indexIn(line) >= 0 )
            {

                // child job
                lineType |= Child;
                child.assign( Job::Id, childRegExp.cap(1) );

            }

        } else if( line.indexOf( "[state ]" ) >= 0 ) {

            if( parentRegExp.indexIn(line) >= 0 )
            {

                // parent ID
                lineType |= Parent;
                job.setParentId( parentRegExp.cap(1).toInt() );

            } else if( resubmissionRegExp.indexIn( line ) >= 0 ) {

                // resubmitted job
                lineType |= Resubmission;
                child.assign( Job::Id, resubmissionRegExp.cap(2) );

            } else if( newJobRegExp.indexIn(line) >= 0 ) {

                // new job
                lineType |= NewJob;
                resubmitted = newJobRegExp.cap(1).toInt();

            }

            // job time stamps
            if( first ) lineType |= Starting;

            if(
                line.indexOf( "to SPLIT" ) >= 0 ||
                line.indexOf( "to ASSIGNED" ) >= 0 ||
                line.indexOf( "to RUNNING" ) >= 0 ||
                line.indexOf( "to STARTED" ) >= 0 ) lineType |= Running;

            else if( line.indexOf( "to ERROR" ) >= 0 || line.indexOf( "to DONE" ) >= 0 ) lineType |= Completed;

        }

        // skip if no valid type set
        if( lineType == None ) continue;

        // parse date
        if( timeRegExp.indexIn(line) < 0 ) continue;

        // parse date
        QString date = QString( "%1 %2 %3:%4" ).arg(timeRegExp.cap(2)).arg(timeRegExp.cap(3)).arg(timeRegExp.cap(4)).arg(timeRegExp.cap(5));

        // assign to job
        first = false;
        if( lineType&Starting ) job.assign( Job::SubmissionTime, date );
        if( lineType&Running ) job.assign( Job::StartTime, date );
        if( lineType&Completed ) job.assign( Job::EndTime, date );

        else if( lineType&(Child|Resubmission) )
        {

            // update children list
            JobSet::iterator iter( children.find( child ) );
            if( iter == children.end() ) {

                // assign submission time and insert in children list
                child.assign( Job::SubmissionTime, date );
                children.insert( child );

            } else {

                // update inserted job
                const_cast<Job*>( &*iter )->assign( Job::SubmissionTime, date );
                if( lineType&Resubmission )
                { const_cast<Job*>( &*iter )->incrementResubmission(); }

            }

        }

    }

    // process job resubmission
    if( resubmitted >= 0 )
    {
        // find previous job from killed jobs list, and reassign parent
        Job sibling = _findKilledJob(  job.get( Job::User ), resubmitted );
        if( sibling.isValid() && sibling.hasParent() )
        { job.setParentId( sibling.parentId() ); }
    }

    // emit signals
    if( emitTrace ) emit jobInformationAvailable( job, stdoutBuffer_.text() );
    else emit jobInformationAvailable( job );

    if( !children.isEmpty() ) emit childrenInformationAvailable( job, children );

    return;
}

//________________________________________________________________________
void AlienInterface::_processJdl( const Job& job )
{
    QRegExp regExp( "^\\s{8,8}" );

    QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
    QString line;
    QString buffer;
    while( !( line = stream.readLine() ).isNull() )
    {
        if( line == "    [" || line == "    ]" ) continue;
        if( line.indexOf( regExp ) < 0 ) continue;
        line.replace( regExp, "" );
        buffer+= line + "\n";
    }

    // find output directory
    QString outputDir;
    QRegExp outputDirRegExp( "OutputDir\\s+=\\s+\\\"(\\S+)\\\"" );
    if( outputDirRegExp.indexIn( buffer ) >= 0 )
    {
        outputDir = outputDirRegExp.cap(1);
        Debug::Throw() << "AlienInterface::_processJdl - output directory: " << outputDir << endl;
    }

    // find output files
    QStringList outputFiles;
    QRegExp outputFilesRegExp( "OutputFile\\s+=\\s+\\{\\s+(\\\"\\S+\\\",*\\s+)+\\}" );
    QRegExp fileRegExp( "\\\"(\\S+)\\\"" );
    if( outputFilesRegExp.indexIn( buffer ) >= 0 )
    {
        // capture strings from matched text
        const QString capturedText( outputFilesRegExp.cap(0) );

        int position = 0;
        while(( position = fileRegExp.indexIn( capturedText, position ) ) >= 0 )
        {
            outputFiles << fileRegExp.cap(1);
            position += fileRegExp.matchedLength();
        }

        Debug::Throw() << "AlienInterface::_processJdl - output files: " << outputFiles.join( " " ) << endl;

    }

    // TODO:
    // - add HTML markups to buffer;
    // - build output files and pass to JdlAvailable
    emit jdlAvailable( job, buffer );
}

//___________________________________________________________________
void AlienInterface::_readStdout(  void )
{
    process_->setReadChannel( QProcess::StandardOutput );
    IOString buffer( *process_ );

    // replace dummy characters
    const QRegExp regExp( "(\\S){2,2}((\\d{2,2};\\d{2,2})|(0))m" );
    buffer.replace( regExp, "" );

    emit logAvailable( buffer );
    stdoutBuffer_.append( buffer );
}

//___________________________________________________________________
void AlienInterface::_readStderr(  void )
{
    process_->setReadChannel( QProcess::StandardError );
    IOString buffer( *process_ );
    emit logAvailable( buffer, Format::Italic );
    stderrBuffer_.append( buffer );
}

//_______________________________________
void AlienInterface::_updateConfiguration( void )
{
    Debug::Throw( "AlienInterface::_updateConfiguration.\n" );

   // get system environment
    systemEnvironment_ = QProcessEnvironment::systemEnvironment();

    // get certificates and check
    const File certificates( XmlOptions::get().raw( "CERTIFICATES_PATH" ) );
    if( certificates.exists() && certificates.isDirectory() )
    {
        systemEnvironment_.insert( "XrdSecSSLCADIR", certificates );
        systemEnvironment_.insert( "X509_CERT_DIR", certificates );
    }

    // get gshell root path from gbbox binary
    const QString gbboxCommand( XmlOptions::get().raw( "GBBOX" ) );
    const QString match( "/bin/gbbox" );
    const int position( gbboxCommand.lastIndexOf( match ) );
    File gShellPath;
    if( position == gbboxCommand.length() - match.length() )
    { gShellPath = gbboxCommand.left( position ) + "/"; }

    // check
    if( gShellPath.exists() && gShellPath.isDirectory() )
    {

        // Store GSHELL_ROOT
        systemEnvironment_.insert( "GSHELL_ROOT", gShellPath );

        // update LD_LIBRARY_PATH if needed
        const QString alienLibPath( gShellPath+"/lib" );
        QString ldPath;
        if( systemEnvironment_.contains( "LD_LIBRARY_PATH" ) )
        { ldPath = systemEnvironment_.value( "LD_LIBRARY_PATH" ); }

        if( ldPath.indexOf( alienLibPath ) < 0 )
        {
            ldPath = ldPath.isEmpty() ? alienLibPath : (alienLibPath + ":" +ldPath);
            systemEnvironment_.insert( "LD_LIBRARY_PATH", ldPath );
        }

    }

}

//____________________________________________________________
void AlienInterface::_resetBuffers( void )
{
    Debug::Throw( "AlienInterface::_resetBuffer.\n" );
    stdoutBuffer_.clear();
    stderrBuffer_.clear();
}

//____________________________________________________________
void AlienInterface::_resetProcess( void )
{
    if( process_ )
    {
        process_->disconnect();
        process_->deleteLater();
        process_ = 0;
    }
}

//____________________________________________________________
CustomProcess* AlienInterface::_newProcess( void )
{

    Debug::Throw( "AlienInterface::_newProcess.\n" );
    _resetProcess();

    // create new process and assign customized environment
    process_ = new CustomProcess( this );
    process_->setProcessEnvironment( systemEnvironment_ );

    connect( process_, SIGNAL(error(QProcess::ProcessError)), SLOT(_error(QProcess::ProcessError)) );
    connect( process_, SIGNAL(finished(int,QProcess::ExitStatus)), this, SLOT(_commandCompleted(int,QProcess::ExitStatus)) );
    connect( process_, SIGNAL(readyReadStandardOutput()), SLOT(_readStdout()) );
    connect( process_, SIGNAL(readyReadStandardError()), SLOT(_readStderr()) );
    return process_;

}

//____________________________________________________________
bool AlienInterface::_isError( const QString& value ) const
{
    if( value.indexOf( "verify failed" ) >= 0 ) return true;
    else if( value.indexOf( "command failed" ) >= 0 ) return true;
    else if( value.indexOf( "err -1" ) >= 0 ) return true;
    else if( value.indexOf( "err 10" ) >= 0 ) return true;
    else return false;
}

//____________________________________________________________
Job AlienInterface::_findKilledJob( const QString& user, int id ) const
{
    JobMap::const_iterator iter( killedJobs_.find( user ) );
    if( iter == killedJobs_.end() ) return Job();

    Job local;
    local.assign( Job::Id, QString::number( id ) );
    JobSet::const_iterator jobIter( iter.value().find( local ) );
    return jobIter == iter.value().end() ? Job():*jobIter;
}

//_______________________________________
bool AlienInterface::_ignoreError( const AlienCommand& command, const QString& message ) const
{

    // check connection state
    if( !( connected_ || command.type() == AlienCommand::Connect ) ) return false;

    return false;

}
