#ifndef Job_h
#define Job_h

// $Id: Job.h,v 1.56 2014/03/21 11:50:54 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "Debug.h"
#include "JobStatistics.h"
#include "TimeStamp.h"
#include "QOrderedSet.h"

#include <QDomDocument>
#include <QDomElement>
#include <QString>
#include <QMap>
#include <QVector>
#include <QHash>

#include <cmath>

//! forward declaration
class JobSet;

//! job
class Job: public Counter
{

    public:

    //! allows fast, keyed access to any data field
    enum FieldType
    {
        Id,
        User,
        Status,
        Command,
        Resubmitted,
        SubmissionTime,
        StartTime,
        EndTime,
        nFields
    };

    //! initialize field names
    static void installFields( void );

    //! install conversions
    static void installConversions( void );

    //! get field name
    static const QString& fieldName( int index )
    { return _fieldNames()[index]; }

    //! empty constructor
    Job( void ):
        Counter( "Job" ),
        parentId_( -1 ),
        fields_( nFields )
    {}

    //: constructor from dom
    explicit Job( const QDomElement& );

    //! constructor from string to browse
    Job( const QString& );

    //! destructor
    virtual ~Job( void )
    {}

    //! dom element
    QDomElement domElement( QDomDocument& ) const;

    //! lower than operator
    bool operator < ( const Job& other ) const
    { return id() < other.id(); }

    //! equal to operator
    bool operator == (const Job& other ) const
    { return id() == other.id(); }

    //! equal to operator
    bool operator != (const Job& other ) const
    { return !(*this == other); }

    //!@name modifiers
    //@{

    //! parent id
    Job& setParentId( const int& value )
    {
        parentId_ = value;
        return *this;
    }

    //! children statistics
    Job& setChildrenStatistics( const JobStatistics& value )
    {
        childrenStatistics_ = value;
        return *this;
    }

    //! set field
    Job& assign( int, const QString& );

    //! set field
    template <typename T>
        Job& assign( int index, const T& value )
    {
        QString stringValue;
        {
            QTextStream stream( &stringValue, QIODevice::WriteOnly );
            stream << value;
        }
        return assign( index, stringValue );
    }

    //! increment resubmission
    void incrementResubmission( void );

    //@}

    //!@name accessors
    //@{

    //! valid
    bool isValid( void ) const
    {
        bool error( false );
        get<int>(Id, &error );
        return (!error && has( User ));
    }

    //! id
    int id( void ) const
    { return has(Id) ? std::abs(get<int>(Id)):-1; }

    //! parent id
    int parentId( void ) const
    { return parentId_; }

    //! returns true job is top level
    bool isMasterJob( void ) const
    { return !hasParent(); }

    //! returns true if job has a parent
    bool hasParent( void ) const
    { return parentId() > 0; }

    //! returns true if has children
    bool hasChildren( void ) const
    { return childrenStatistics_.get( JobStatistics::Total ) > 0; }

    //! children statistics
    const JobStatistics& childrenStatistics( void ) const
    { return childrenStatistics_; }

    //! returns true if this job is a child of argument
    bool isChild( const Job& job ) const
    { return parentId() == job.id(); }

    //! true if inserting
    bool isInserting( void ) const;

    //! true if running
    bool isRunning( void ) const;

    //! true if completed
    bool isCompleted( void ) const;

    //! true if pending
    bool isPending( void ) const;

    //! true if has error
    bool isFailed( void ) const;

    //! true if job is killed
    bool isKilled( void ) const;

    //! true if job might have children (based on status)
    bool mightHaveChildren( void ) const;

    //! return status
    QString status( void ) const;

    //! return error code
    QString error( void ) const;

    //! const field accessor
    QString get( int index ) const;

    //! field accessor (cast to anything)
    template <typename T>
        T get( int index, bool* error = 0 ) const
    {
        T out;
        QString value( get( index ) );
        QTextStream in( &value, QIODevice::ReadOnly );
        in >> out;
        if( in.status() != QTextStream::Ok )
        {
            if( error ) *error = true;
            return T();
        } else return out;
    }

    //! Check if field makes sence
    bool has( int index ) const
    { return get(index).size(); }

    //! resubmission count
    QString resubmitted( void ) const;

    //! times
    const TimeStamp& submissionTime( void ) const
    { return submissionTime_; }

    //! times
    const TimeStamp& startTime( void ) const
    { return startTime_; }

    //! times
    const TimeStamp& endTime( void ) const
    { return endTime_; }

    //@}

    //! used to count total number of jobs
    class TotalFTor
    {
        public:

        //! constructor
        TotalFTor( void )
        {}

        //! destructor
        virtual ~TotalFTor( void )
        {}

        //! predicate
        bool operator() (const Job& job ) const
        {  return !( job.isMasterJob() && job.hasChildren() ); }

    };

    //! used to select pending jobs
    class PendingFTor
    {
        public:

        //! constructor
        PendingFTor( void )
        {}

        //! destructor
        virtual ~PendingFTor( void )
        {}

        //! predicate
        bool operator() (const Job& job ) const
        {  return job.isPending() && !job.hasChildren(); }

    };

    //! used to select running jobs
    class RunningFTor
    {
        public:

        //! constructor
        RunningFTor( void )
        {}

        //! destructor
        virtual ~RunningFTor( void )
        {}

        //! predicate
        bool operator() (const Job& job ) const
        {  return job.isRunning() && !job.hasChildren(); }

    };

    //! used to select completed jobs
    class CompletedFTor
    {
        public:

        //! constructor
        CompletedFTor( void )
        {}

        //! predicate
        bool operator() (const Job& job ) const
        {  return job.isCompleted() && !job.hasChildren(); }

    };

    //! used to select failed jobs
    class FailedFTor
    {
        public:

        //! constructor
        FailedFTor( void )
        {}

        //! destructor
        virtual ~FailedFTor( void )
        {}

        //! predicate
        bool operator() (const Job& job ) const
        {  return job.isFailed() && !job.hasChildren(); }

    };

    //! used to select jobs with given Id
    class SameIdFTor
    {

        public:

        //! constructor
        SameIdFTor( const int& id ):
            id_( id )
        {}

        //! predicate
        bool operator() (const Job& job ) const
        {  return job.id() == id_; }

        private:

        //! field index
        int id_;

    };

    //! typedef for list of jobs
    typedef int id_t;
    typedef QList<Job> List;
    class IdSet: public QOrderedSet<id_t>, public Counter
    {
        public:

        //! default constructor
        IdSet( void ):
            Counter( "Job::IdSet" )
        {}

        //! constructor from JobSet
        IdSet( const JobSet& );

        //! constructor from list
        IdSet( const Job::List& jobs);

    };

    protected:

    //! set field id from xml field name return false if not found
    static bool _xmlToId( const QString& name, int& id )
    {
        QMap<QString, unsigned int>::iterator iter = _fieldMap().find( name );
        if( iter == _fieldMap().end() ) return false;
        id = iter.value();
        return true;
    }

    //! fields
    typedef QVector<QString> FieldVector;

    //! field names
    static FieldVector& _fieldNames( void );

    //! status conversions
    typedef QHash<QString,QString> ConversionMap;
    static ConversionMap& _conversions( void );

    //! load field from DomElement
    void _loadXmlField( const QDomElement& element );

    //! map between fieldNames and field_id, for fast reading access
    typedef QMap<QString, unsigned int > FieldMap;

    //! xml field map
    static FieldMap& _fieldMap( void );

    //! set xml_field name for given index
    static void _setFieldName( FieldType index, const QString& value )
    {
        _fieldNames()[index] = value;
        _fieldMap().insert( value, index );
    }

    // convert string to time
    TimeStamp _time( const QString& );

    private:

    //! parent id
    /*! it is kept outside of fields in order not to appear in list */
    int parentId_;

    //! fields
    FieldVector fields_;

    //! children statistics
    JobStatistics childrenStatistics_;

    // timestamps
    TimeStamp submissionTime_;
    TimeStamp startTime_;
    TimeStamp endTime_;
};

//! hash
inline unsigned int qHash( const Job& job )
{ return qHash( job.id() ); }

#endif
